import * as puppeteer from "puppeteer";
import { test, beforeEach, afterEach } from "tap";
import buildIndex from "../scripts/buildIndex";
const launchArgs = process.env.CI === "true" ? ["--no-sandbox"] : [];
let browser: puppeteer.Browser;
let page: puppeteer.Page;

beforeEach(async () => {
  const result = await buildIndex({
    write: false,
    sourcemap: false,
    format: "esm",
  });
  const [file] = result.outputFiles;
  const code = Buffer.from(file.contents);
  browser = await puppeteer.launch({
    args: [...launchArgs, "--disable-web-security"],
  });
  page = await browser.newPage();
  await page.setRequestInterception(true);
  page.on("request", (req) => {
    if (req.url() === "https://fake.test/") {
      return req.respond({
        body: '<iframe sandbox="allow-scripts"></iframe>',
        contentType: "text/html",
      });
    }

    if (req.url().match(/pmrpc.js$/)) {
      return req.respond({
        body: code,
        contentType: "application/javascript",
      });
    }
    req.abort();
  });
  await page.goto("https://fake.test");
  await page.addScriptTag({
    type: "module",
    content: `
    import JsonRpc from './pmrpc.js'
    let resolve;
    let reject;
    const handshakePromise = new Promise(r => resolve = r);
    self.handshakePromise = handshakePromise;
    self.addEventListener('message', e => {
      const [port] = e.ports;
      port.start();
      self.rpc = new JsonRpc({source: port, destination: port});
      resolve();
    });
  `,
  });

  const [, frame] = page.frames();
  await frame.addScriptTag({
    type: "module",
    content: `
    import JsonRpc from './pmrpc.js';
    const {port1, port2} = new MessageChannel();

    const rpc = new JsonRpc({
      source: port1,
      destination: port1,
      methods: {
        greet: name => 'Hello, ' + name,
      },
    });
    port1.start();

    window.top.postMessage({}, '*', [port2]);
  `,
  });
  // await page.waitFor(100);
  await page.evaluate(`window.handshakePromise`);
});

afterEach(async () => {
  (await browser && browser.close());
});

test("calls between iframes", async (t) => {
  const result = await page.evaluate(`self.rpc.call('greet', 'Alice')`);
  t.same(result, "Hello, Alice");
});

test('fails on missing', async t => {
  await t.rejects(page.evaluate(`self.rpc.call('missing')`, {
    message: /method not found/i,
  }));
});
