import JsonRpc, {ErrorCodes} from './JsonRpc';
import {test} from 'tap';
import {spy} from 'sinon';

const defer = (ms: number, val?: any) =>
  new Promise(resolve => setTimeout(() => resolve(val), ms));

const buildRequest = (method: string, params?: any[] | object) => ({
  jsonrpc: '2.0',
  id: 123,
  method,
  params,
});

test('handling requests', async ({beforeEach, test}) => {
  let rpc: any;

  beforeEach(() => {
    rpc = new JsonRpc({
      methods: {
        hello: (name: string) => `goodbye ${name}`,
        canVote: (voter: {age: number}) => voter.age > 18,
        deferred: () => Promise.resolve('deferredResult'),
        blowUp: () => {
          throw Error('BOOM!');
        },
      },
    });
  });

  test('handles array params', async t => {
    const request = buildRequest('hello', ['Alice']);
    const response = await rpc.handleRequest(request);

    t.same(response.jsonrpc, '2.0');
    t.same(response.id, request.id);
    t.same(response.result, 'goodbye Alice');
  });

  test('handles object params', async t => {
    const request = buildRequest('canVote', {age: 22});
    const response = await rpc.handleRequest(request);

    t.same(response.jsonrpc, '2.0');
    t.same(response.id, request.id);
    t.same(response.result, true);
  });

  test('waits on promise results', async t => {
    const request = buildRequest('deferred');
    const response = await rpc.handleRequest(request);

    t.same(response.result, 'deferredResult');
  });

  test('errors on missing method', async t => {
    const request = buildRequest('missing');
    const response = await rpc.handleRequest(request);

    t.same(response.id, request.id);
    t.same(response.error.code, ErrorCodes.MethodNotFound);
  });

  test('errors on throwing method', async t => {
    const request = buildRequest('blowUp');
    const response = await rpc.handleRequest(request);

    t.same(response.id, request.id);
    t.same(response.error.code, ErrorCodes.InternalError);
  });
});

test('invalid messages', async ({beforeEach, test}) => {
  let rpc: any;
  let messageEvent : any;

  beforeEach(() => {
    rpc = new JsonRpc({
      methods: {
        spy: spy(),
      },
    });
    messageEvent = {
      origin: '*',
      data: buildRequest('spy')
    };
  });

  test('ignores messages missing data', async t => {
    delete messageEvent.data;
    await rpc.handleMessage(messageEvent);
    t.notOk(rpc.methods.spy.called);
  });

  test('ignores requests missing jsonrpc', async t => {
    delete messageEvent.data.jsonrpc;
    await rpc.handleMessage(messageEvent);
    t.notOk(rpc.methods.spy.called);
  });

  test('ignores requests with wrong origin', async t => {
    rpc.origin = 'http://example.com';
    messageEvent.origin = 'http://fake.com';
    await rpc.handleMessage(messageEvent);
    t.notOk(rpc.methods.spy.called);
  });
});

test('mounted', async ({beforeEach, afterEach, test}) => {
  let frame1: HTMLIFrameElement;
  let frame2: HTMLIFrameElement;
  let rpc1: JsonRpc;
  let rpc2: JsonRpc;
  let document: Document;

  beforeEach(async () => {
    const {JSDOM} = await import('jsdom');
    document = new JSDOM().window.document;
    frame1 = document.createElement('iframe');
    frame2 = document.createElement('iframe');
    document.body.appendChild(frame1);
    document.body.appendChild(frame2);

    rpc1 = new JsonRpc({
      methods: {},
      source: frame1.contentWindow as MessageEventSource,
      destination: frame2.contentWindow as MessageEventSource,
    });

    rpc2 = new JsonRpc({
      destination: frame1.contentWindow as MessageEventSource,
      source: frame2.contentWindow as MessageEventSource,
      methods: {
        greet: (name: string) => `Hello, ${name}`,
        explode: () => {
          throw Error('Kapow');
        },
      },
    });
  });

  afterEach(() => {
    document.body.removeChild(frame1);
    document.body.removeChild(frame2);
  });

  test('communicates between iframes', async t => {
    const result = await rpc1.call('greet', ['Alice']);
    t.same(result, `Hello, Alice`);
  });

  test('propagates errors between iframes', async t => {
    await t.rejects(rpc1.apply('explode'), {
      code: ErrorCodes.InternalError,
    });
  });

  test('propagates missing between iframes', async t => {
    await t.rejects(rpc1.apply('missing'), {
      code: ErrorCodes.MethodNotFound,
    });
  });

  test('unmounts', async t => {
    rpc1.unmount();
  });

  test('ignores unexpected responses', async t => {
    const {contentWindow} = frame1;
    contentWindow!.postMessage({jsonrpc: '2.0', result: {}}, '*');
    await defer(0);
  });

  test('ignores responding without destination', async t => {
    delete rpc2['destination'];
    const result = await Promise.race([
      defer(10, 'timeout'),
      rpc1.call('greet', 'Norah'),
    ]);
    t.same(result, 'timeout');
  })
});

test('errors when applying without a destination', async t => {
  const rpc = new JsonRpc();
  t.throws(() => rpc.apply('yo'));
});

test('does not pass origin to postMessage when not window', async t => {
  const postMessage = spy();
  const destination = {postMessage} as unknown as MessageEventSource;
  const rpc = new JsonRpc({destination});
  rpc.call('x');

  t.notOk(postMessage.calledWith(null, undefined))
});